from multiprocessing import Process, pool

class NoDaemonProcess(Process):
    """A process with the deamon property set to False.
    
        Purpose: multiprocessing.pool.Pool makes processes daemonic and then immediately starts them.
                 It's not possible to re-set their daemon attribute to False before they are 
                 started and afterwards it's not allowed anymore. This circumvents that.
    """
    # make 'daemon' attribute always return False
    def _get_daemon(self):
        return False
    def _set_daemon(self, value):
        pass
    daemon = property(_get_daemon, _set_daemon)
 
# We sub-class multiprocessing.pool.Pool instead of multiprocessing.Pool
# because the latter is only a wrapper function, not a proper class.
class NoDaemonProcessPool(pool.Pool):
    Process = NoDaemonProcess

# taken from https://stackoverflow.com/questions/6974695/python-process-pool-non-daemonic


import os
import sys
spychronous_dir = os.path.dirname(os.path.dirname(os.path.realpath(__file__)))
sys.path += [spychronous_dir]
from spychronous import SynchronousJob

import logging
logger = logging.getLogger('spychronous')
logger.disabled = True

def run_forever_with_item(item):
    while True:
        pass
    return item

if __name__ == '__main__':
    items = [1, 2, 3]
    running_forever_job = SynchronousJob(func=run_forever_with_item, items=items)
    running_forever_job.run_multi_processed()
    print 'we will never get here'


import os
import sys
spychronous_dir = os.path.dirname(os.path.dirname(os.path.realpath(__file__)))
sys.path += [spychronous_dir]
from spychronous import SynchronousJob, run_function
import unittest
import multiprocessing
import subprocess
import shlex
import signal
from time import sleep

import logging
logger = logging.getLogger('spychronous')
logger.disabled = True

class TestRun(object):
    maxDiff = None

    def setUp(self):
        self.addend = 2
        self.items = [1,2,3]
        self.items_without_3 = [i != 3 and i or None for i in self.items]
        self.instance_method_name = None

    def test_worker_return_values_are_aggregated(self):
        plus_num_job = SynchronousJob(func=get_plus_num, items=self.items, args=[self.addend])
        new_items = getattr(plus_num_job, self.instance_method_name)()
        items_plus_addend = map(lambda i: i+self.addend, self.items)
        self.assertEqual(new_items, items_plus_addend)

        # test that no values are collected when SynchronousJob fails.
        zero_div_error_job = SynchronousJob(func=perform_ZeroDivisionError, items=self.items)
        with self.assertRaises(ZeroDivisionError):
            output = getattr(zero_div_error_job, self.instance_method_name)()
        with self.assertRaises(UnboundLocalError): # output was never able to be defined.
            output

        # test outputs are collected when worker exceptions aren't raised.
        allowing_number_except_3_job = SynchronousJob(func=get_number_except_perform_ZeroDivisionError_on_3, items=self.items, suppress_worker_exceptions=True)
        output = getattr(allowing_number_except_3_job, self.instance_method_name)()
        self.assertEqual(output, self.items_without_3)
        
    def test_suppress_raised_worker_exceptions_and_complete_execution(self):
        # testing the opposite first.
        zero_div_error_job = SynchronousJob(func=perform_ZeroDivisionError, items=self.items, suppress_worker_exceptions=False)
        with self.assertRaises(ZeroDivisionError):
            getattr(zero_div_error_job, self.instance_method_name)()

        # testing correct behavior.
        zero_div_error_job = SynchronousJob(func=get_number_except_perform_ZeroDivisionError_on_3, items=self.items, suppress_worker_exceptions=True)
        output = getattr(zero_div_error_job, self.instance_method_name)()
        self.assertEqual(output, self.items_without_3)

class TestRunMultiProcessed(TestRun, unittest.TestCase):
    def setUp(self):
        super(TestRunMultiProcessed, self).setUp()
        self.instance_method_name = 'run_multi_processed'

    def test_handle_sigint(self):
        child_program_name = 'run_forever.py'
        # Kill any processes with run_forever.py running.
        process_records = run_command('ps -A')
        for record in process_records:
            if child_program_name in record:
                pid = int(line.split(None, 1)[0])
                os.kill(pid, signal.SIGKILL)
        # Spin up the never-self-terminating subprocess.
        python_interpreter = sys.executable # user's current python interpreter.
        run_forever_script_path = os.path.join(spychronous_dir, 'tests', 'run_forever.py')
        cmd = python_interpreter + ' ' + run_forever_script_path
        process = subprocess.Popen(shlex.split(cmd), stderr=subprocess.PIPE) # stderr will contain sigint -- we don't want to see it.
        # Verify that the subprocess exists and has spawned children.
        greppable_path = get_greppable(run_forever_script_path)
        exact_interpreter = run_command("ps -ef | grep '%s' | awk '{ print $8 }'" % greppable_path).splitlines()[0] # the python path could have been expanded in the subprocess.
        exact_cmd = exact_interpreter + ' ' + run_forever_script_path
        greppable_cmd = get_greppable(exact_cmd)
        parent_pid = run_command("ps -ef | grep '%s' | awk '{ print $2 }'" % greppable_cmd).splitlines()[0]
        if not parent_pid:
            process.send_signal(signal.SIGINT)
            raise Exception('No parent pid!!!')
        greppable_pid = get_greppable(parent_pid)
        children_pids = run_command("ps -ef | grep '%s' | awk '{ print $2 }' | grep -v '%s'" % (greppable_pid, greppable_pid)).splitlines()
        self.assertGreater(len(children_pids), 0)
        # Kill the subprocess.
        process.send_signal(signal.SIGINT)
        WAIT_PROCESS_DIE_SECS = 1
        sleep(WAIT_PROCESS_DIE_SECS)
        # Verify we've killed the subprocess and its children (the children's routine runs forever -- if they aren't dead, sigint didn't propogate properly).
        self.assertTrue(process.poll())
        WAIT_CHILDREN_DIE_SECS = 1
        sleep(WAIT_CHILDREN_DIE_SECS)
        children_pids = run_command("ps -ef | grep '%s' | awk '{ print $2 }' | grep -v '%s'" % (greppable_pid, greppable_pid)).splitlines()
        self.assertEqual(len(children_pids), 0)

class TestRunSingleProcessed(TestRun, unittest.TestCase):
    def setUp(self):
        super(TestRunSingleProcessed, self).setUp()
        self.instance_method_name = 'run_single_processed'

class TestRunFunction(unittest.TestCase):
    maxDiff = None

    def setUp(self):
        self.input_args = [1]
        self.output_nums = [] # we collect the output and build upon it as we proceed.

    def test_run_function(self):
        input_args = self.input_args
        output_nums = self.output_nums
        run_function(get_plus_one, input_args, output_nums)
        self.assertEqual(output_nums, [2])

        input_args = [5]
        run_function(get_plus_one, input_args, output_nums)
        self.assertEqual(output_nums, [2, 6])

        with self.assertRaises(ZeroDivisionError):
            run_function(perform_ZeroDivisionError, input_args, output_nums)
        self.assertEqual(output_nums, [2, 6, None])


### SIMPLE FUNCTIONS TO USE FOR TESTING ###
def get_plus_one(number):
    return number + 1

def get_plus_num(number1, number2):
    return number1 + number2

def perform_ZeroDivisionError(number):
    1/0

def get_number_except_perform_ZeroDivisionError_on_3(number):
    if number == 3:
        perform_ZeroDivisionError(number)
    return number

### HELPER FUNCTIONS ###
def run_command(cmd):
    if "|" in cmd:
        cmd_parts = cmd.split('|')
    else:
        cmd_parts = [cmd]
    p = {}
    for i, cmd_part in enumerate(cmd_parts):
        cmd_part = cmd_part.strip()
        if i == 0:
            p[i]=subprocess.Popen(shlex.split(cmd_part),stdin=None, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        else:
            p[i]=subprocess.Popen(shlex.split(cmd_part),stdin=p[i-1].stdout, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    (stdout_data, std_error_data) = p[i].communicate()
    exit_code = p[0].wait()

    if exit_code != 0:
        print "Output:"
        print str(stdout_data)
        raise Exception(std_error_data)
    else:
        return str(stdout_data).strip()

def get_greppable(string):
    """Simply produces a string that -- when grepped -- will omit listing the grep process in a grep listing.
    """
    return string.replace(string[0], '[%s]' % string[0], 1)


if __name__ == '__main__':
    unittest.main(verbosity=2)



from setuptools import setup, find_packages

version = '1.0.post4'

with open('README.md', 'r') as fh:
  long_description = fh.read()

setup(
  name = 'spychronous',
  py_modules = ['spychronous'],
  packages=find_packages('src'),
  package_dir={'':'src'},
  version = version,
  license='MIT',
  description = 'A simple synchronous job runner for parallel processing tasks in Python.',
  long_description = long_description,
  long_description_content_type = 'text/markdown',
  author = 'Santiago C Paredes',
  author_email = 'santiago.paredes2012@gmail.com',
  url = 'https://github.com/scparedes/spychronous',
  download_url = 'https://github.com/scparedes/spychronous/releases/download/%s/spychronous-%s.tar.gz' % (version, version),
  keywords = ['multiprocessing', 'multiprocess', 'multi', 'process', 'synchronous', 'job', 'runner', 'simple'],
  install_requires=[
  ],
  classifiers=[
    'Development Status :: 5 - Production/Stable',
    'Intended Audience :: Developers',
    'Topic :: Software Development :: Build Tools',
    'License :: OSI Approved :: MIT License',
    'Programming Language :: Python :: 2',
    'Programming Language :: Python :: 2.6',
    'Programming Language :: Python :: 2.7',
  ],
)



from multiprocessing import Pool, Manager
from non_daemonic_process.non_daemonic_processing import NoDaemonProcessPool
import sys
import signal

import traceback
import logging

LOG = logging.getLogger(__name__)

HOURS = 60*60
DAYS = HOURS*24
DEFAULT_TIMEOUT = 30 * DAYS

class SynchronousJob(object):
    """A synchronous Job-runner that leverages parallel processing to apply a function to each item in a list.
        Args:
            func (function): The function that's applied to each item -- the first parameter must represent a single item from items.
            items (list): The dataset that's iterated over and transformed with func.
            args (list): The additional parameters in func's signature proceeding the first required parameter.
            processes (int):The number of processes you want to enlist for parallelization.
            timeout (int): The number of seconds a given process in the process pool has to complete its work.
            no_daemon (bool): This allows processes in the process pool to spawn more pools of processes.
            suppress_worker_exceptions (bool): Prevents killing a SynchronousJob and its workers when any given coworker raises an exception.
    """
    def __init__(self, func=None, items=[], args=[], processes=4, timeout=DEFAULT_TIMEOUT, no_daemon=False, suppress_worker_exceptions=False):
        self.func = func
        self.items = items
        self.args = args
        self.processes = processes
        self.timeout = timeout
        self.no_daemon = no_daemon
        self.suppress_worker_exceptions = suppress_worker_exceptions
    
    def run_single_processed(self, log_start_finish=False):
        if log_start_finish:
            LOG.info('Beginning single-processed SynchronousJob...')
        worker_outputs = list() # This will accumulate return values of 'func'.
        for item in self.items:
            try:
                run_function(self.func, [item] + self.args, worker_outputs)
            except Exception as e:
                if not self.suppress_worker_exceptions:
                    raise e
                else:
                    LOG.info("Logging '%s:%s' but neglecting to raise it" % (e.__class__.__name__, e.message))
        if log_start_finish:
            LOG.info('Finished single-processed SynchronousJob...')
        return worker_outputs

    def run_multi_processed(self, log_start_finish=False):
        if log_start_finish:
            LOG.info('Beginning multi-processed SynchronousJob...')
        # The correct way to handle Ctrl+C/SIGINT with multiprocessing.Pool is to:
        # 1) Make the process ignore SIGINT before a process Pool is created. 
        #    This way created child processes inherit SIGINT handler.
        # 2) Restore the original SIGINT handler in the parent process after a Pool has been created.
        # 3) Wait on the results with timeout because the default "blocking-waits" ignore all signals.
        # ** Based on https://stackoverflow.com/a/35134329/3577492

        # SIGINT-handling step 1
        original_sigint_handler = signal.signal(signal.SIGINT, signal.SIG_IGN)

        manager = Manager()
        worker_outputs = manager.list() # This will accumulate return values of 'func'.

        pool_params = {'processes':self.processes,
                       'maxtasksperchild':1} # In tandum with timeout, this implements the process timeout.
        if sys.version_info[0] < 2.7:
            pool_params.pop('maxtasksperchild')

        # SIGINT-handling step 2
        if self.no_daemon:
            pool = NoDaemonProcessPool(**pool_params)
        else:
            pool = Pool(**pool_params)
        signal.signal(signal.SIGINT, original_sigint_handler)

        worker_statuses = [] # A list of AsyncResult's.
        for item in self.items:
            worker_statuses.append(pool.apply_async(run_function, [self.func, [item] + self.args, worker_outputs]))
        pool.close() # no more work will be submitted to workers

        for ws in worker_statuses:
            try:
                # SIGINT-handling step 3
                ws.get(self.timeout)  # checks for worker errors
            except KeyboardInterrupt:
                LOG.info('caught KeyboardInterrupt, terminating workers')
                pool.terminate()
                raise
            except Exception as e:
                if not self.suppress_worker_exceptions:
                    pool.terminate()
                    raise e
                else:
                    LOG.info("Logging '%s:%s' but neglecting to raise it" % (e.__class__.__name__, e.message))

        pool.join() # wait for worker processes to terminate
        if log_start_finish:
            LOG.info('Finished multi-processed SynchronousJob...')
        return list(worker_outputs)

def run_function(some_function, args, worker_outputs):
    """This is a method that will be called by every SynchronousJob worker.
        Args:
            some_function (function): The function that will be applied to 'args'.
            args (list): The complete list of args that supplies some_function's signature.
            worker_outputs (list): The return values of any given some_function result appeneded to this list.
    
        Notes:
            - Stacktraces from exceptions are logged so users can see exactly why a worker failed.
            - Multiprocess note: worker_outputs is a multiprocessing.managers.ListProxy for multi_processed SynchronousJob runs.
    """
    try:
        output = some_function(*args)
        worker_outputs.append(output)
    except Exception as e:
        worker_outputs.append(None)
        stacktrace = traceback.format_exc()
        LOG.error(stacktrace)
        raise e
